"use strict";

function _interopDefault(ex) {
  return ex && "object" == typeof ex && "default" in ex ? ex.default : ex;
}

Object.defineProperty(exports, "__esModule", {
  value: !0
});

var _objectSpread = _interopDefault(require("@babel/runtime/helpers/objectSpread")), semver = _interopDefault(require("semver"));

function getDependents(releases, workspaces, dependencyGraph) {
  let updated = !1, pkgsToSearch = [ ...releases ], pkgJsonsByName = new Map(workspaces.map(({name: name, config: config}) => [ name, config ]));
  for (;pkgsToSearch.length > 0; ) {
    const nextRelease = pkgsToSearch.shift();
    if (!nextRelease) continue;
    const pkgDependents = dependencyGraph.get(nextRelease.name);
    if (!pkgDependents) throw new Error(`Error in determining dependents - could not find package in repository: ${nextRelease.name}`);
    pkgDependents.map(dependent => {
      let type;
      const dependentPkgJSON = pkgJsonsByName.get(dependent);
      if (!dependentPkgJSON) throw new Error("Dependency map is incorrect");
      const {depTypes: depTypes, versionRange: versionRange} = getDependencyVersionRange(dependentPkgJSON, nextRelease.name);
      return !depTypes.includes("peerDependencies") || "patch" === nextRelease.type || releases.some(dep => dep.name === dependent) && !releases.some(dep => dep.name === dependent && "major" !== dep.type) ? releases.some(dep => dep.name === dependent) || semver.satisfies(semver.inc(nextRelease.oldVersion, nextRelease.type), versionRange) || (type = "patch") : type = "major", 
      {
        name: dependent,
        type: type,
        pkgJSON: dependentPkgJSON
      };
    }).filter(({type: type}) => type).forEach(({name: name, type: type, pkgJSON: pkgJSON}) => {
      updated = !0;
      const existing = releases.find(dep => dep.name === name);
      if (existing && "major" === type && "major" !== existing.type) existing.type = "major", 
      pkgsToSearch.push(existing); else {
        let newDependent = {
          name: name,
          type: type,
          oldVersion: pkgJSON.version,
          changesets: []
        };
        pkgsToSearch.push(newDependent), releases.push(newDependent);
      }
    });
  }
  return updated;
}

function getDependencyVersionRange(dependentPkgJSON, dependencyName) {
  const DEPENDENCY_TYPES = [ "dependencies", "devDependencies", "peerDependencies", "optionalDependencies" ], dependencyVersionRange = {
    depTypes: [],
    versionRange: ""
  };
  for (const type of DEPENDENCY_TYPES) {
    const deps = dependentPkgJSON[type];
    deps && (deps[dependencyName] && (dependencyVersionRange.depTypes.push(type), dependencyVersionRange.versionRange = deps[dependencyName]));
  }
  return dependencyVersionRange;
}

function flattenReleases(changesets, workspaces) {
  let releases = new Map();
  return changesets.forEach(changeset => {
    changeset.releases.forEach(({name: name, type: type}) => {
      let release = releases.get(name), ws = workspaces.find(ws => ws.name === name);
      if (!ws) throw new Error(`Could not find package information for ${name}`);
      let {config: config} = ws;
      release ? ("minor" === release.type && "major" === type ? release.type = type : "patch" !== release.type || "major" !== type && "minor" !== type || (release.type = type), 
      release.changesets.push(changeset.id)) : release = {
        name: name,
        type: type,
        oldVersion: config.version,
        changesets: [ changeset.id ]
      }, releases.set(name, release);
    });
  }), [ ...releases.values() ];
}

function applyLinks(releases, workspaces, linked) {
  let updated = !1;
  if (!linked) return updated;
  for (let linkedPackages of linked) {
    let highestReleaseType, highestVersion, releasingLinkedPackages = releases.filter(release => linkedPackages.includes(release.name));
    if (!(releasingLinkedPackages.length < 1)) {
      for (let pkg of releasingLinkedPackages) highestReleaseType ? "major" === pkg.type ? highestReleaseType = pkg.type : "minor" === pkg.type && "major" !== highestReleaseType && (highestReleaseType = pkg.type) : highestReleaseType = pkg.type;
      for (let linkedPackage of linkedPackages) {
        let workspace = workspaces.find(workspace => workspace.name === linkedPackage);
        if (!workspace) throw console.error(`FATAL ERROR IN CHANGESETS! We were unable to version for linked package: ${linkedPackage} in linkedPackages: ${linkedPackages.toString()}`), 
        new Error("fatal: could not resolve linked packages");
        (void 0 === highestVersion || semver.gt(workspace.config.version, highestVersion)) && (highestVersion = workspace.config.version);
      }
      if (!highestVersion || !highestReleaseType) throw new Error("Large internal changesets error in calculating linked versions. Please contact the maintainers");
      for (let linkedPackage of releasingLinkedPackages) linkedPackage.type !== highestReleaseType && (updated = !0, 
      linkedPackage.type = highestReleaseType), linkedPackage.oldVersion !== highestVersion && (updated = !0, 
      linkedPackage.oldVersion = highestVersion);
    }
  }
  return updated;
}

function assembleReleasePlan(changesets, workspaces, dependentsGraph, config) {
  let releases = flattenReleases(changesets, workspaces), releaseObjectValidated = !1;
  for (;!1 === releaseObjectValidated; ) {
    let dependentAdded = getDependents(releases, workspaces || [], dependentsGraph);
    releaseObjectValidated = !applyLinks(releases, workspaces, config.linked) && !dependentAdded;
  }
  return {
    changesets: changesets,
    releases: releases.map(incompleteRelease => _objectSpread({}, incompleteRelease, {
      newVersion: semver.inc(incompleteRelease.oldVersion, incompleteRelease.type)
    }))
  };
}

exports.default = assembleReleasePlan;
